#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <glib/gprintf.h>
#include <GL/gl.h>
#include <sys/stat.h>


#include <config.hpp>



#define _3DFB_CONFIG_DIR ".3dfb"


static const GLfloat default_dir_color[] = {1.0f, 1.0f, 1.0f};
static const GLfloat default_pdir_color[] = {0.0f, 1.0f, 0.0f};
static const GLfloat default_file_color[] = {0.5f, 0.5f, 0.9f};
static const GLfloat default_fog_color[] = {0.0f, 0.0f, 0.0f, 1.0f};

static guint HashFunction (gconstpointer key);
static gint HashCompare (gconstpointer sName1, gconstpointer sName2);
static void set_option_value (gpointer key, gpointer value, gpointer extra_data);



Options::Options()
{
    /*
     * In the future, we will check here for
     * environment variables, like 3DFB_CONFIG_FILE,
     * or system-wide config files like
     * /usr/share/3dfb/config
     */
    config_file = g_strconcat(g_get_home_dir(), G_DIR_SEPARATOR_S,
                              _3DFB_CONFIG_DIR, G_DIR_SEPARATOR_S,
                              "options", 0);
    options_readed = FALSE;
}


Options::~Options()
{
    g_free(config_file);
}


int Options::read_options ()
{
    GError *error;
    gchar *buffer;
    //gchar *error_message;
    GHashTable *options_table;

    /*
     * Check for config file existence
     */
    if (g_file_test(config_file, G_FILE_TEST_EXISTS)) {
        error = NULL;
        g_file_get_contents(config_file, &buffer, NULL, &error);
        if (error) {
            // error_message = g_strdup(error->message);
            g_error_free(error);
            return -1;
        }
        options_table = g_hash_table_new(HashFunction, HashCompare);
        if (parse_options(buffer, options_table) < 0)
            return -1;
        g_hash_table_foreach(options_table, set_option_value, this);
        g_hash_table_destroy(options_table);
        options_readed = TRUE;
    } else {
        set_default_options();
        options_readed = TRUE;
        create_new_config_file();
        save_options();
    }
    return 0;
}


/*
 * Save options to disk.
 *
 * Warning: if 'config_file' already exists, it is overwritten.
 *
 * Retuns 0 on success, -1 on error.
 * In future, this will retrieve an error message too, if any.
 */
int Options::save_options ()
{
    FILE *fp;
    gchar *error_msg;
    gchar *value, *tmp;

    // Check if read_options() has been called
    if (!options_readed) {
        //error_message = g_strdup ("There is no options to save");
        return -1;
    }
    fp = fopen(config_file, "w");
    if (!fp) {
        error_msg = g_strdup_printf("%s: config file %s cannot be created",
                                    PACKAGE_NAME, config_file);
        perror(error_msg);
        g_free (error_msg);
        return -1;
    }

    /*
     * FIXME would be better to write all to a buffer, and then
     * copy the whole buffer to file. That way, on error, we don't
     * get a corrupted file.
     */
    
    // Write directory color
    if (fwrite("dir_color=", 1, strlen("dir_color="), fp) < strlen("dir_color=")) {
        perror("Save options failed");
        return -1;
    }
    if (fwrite("{", 1, 1, fp) < 1)
        return -1;        
    for (int i = 0; i < 3; i++) {
        // .3 precision would be enough?
        value = g_strdup_printf("%.3f", dir_color[i]);
        if (!value)
            return -1;
        if (i < 2)
            tmp = g_strconcat(value, ",", 0);
        else
            tmp = g_strconcat(value, "}", 0);
        g_free(value);
        if (fwrite(tmp, sizeof(gchar), strlen(tmp), fp) < strlen(tmp)) {
            g_free(tmp);
            return -1;
        }
        g_free(tmp);
    }
    if (fwrite("\n", 1, 1, fp) < 1)
        return -1;        

    // Write parent directory color
    if (fwrite("pdir_color=", 1, strlen("pdir_color="), fp) < strlen("pdir_color=")) {
        perror("Save options failed");
        return -1;
    }
    if (fwrite("{", 1, 1, fp) < 1)
        return -1;        
    for (int i = 0; i < 3; i++) {
        // .3 precision would be enough?
        value = g_strdup_printf("%.3f", pdir_color[i]);
        if (!value)
            return -1;
        if (i < 2)
            tmp = g_strconcat(value, ",", 0);
        else
            tmp = g_strconcat(value, "}", 0);
        g_free(value);
        if (fwrite(tmp, sizeof(gchar), strlen(tmp), fp) < strlen(tmp)) {
            g_free(tmp);
            return -1;
        }
        g_free(tmp);
    }
    if (fwrite("\n", 1, 1, fp) < 1)
        return -1;        

    // Write file color
    if (fwrite("file_color=", 1, strlen("file_color="), fp) < strlen("file_color=")) {
        perror("Save options failed");
        return -1;
    }
    if (fwrite("{", 1, 1, fp) < 1)
        return -1;        
    for (int i = 0; i < 3; i++) {
        // .3 precision would be enough?
        value = g_strdup_printf("%.3f", file_color[i]);
        if (!value)
            return -1;
        if (i < 2)
            tmp = g_strconcat(value, ",", 0);
        else
            tmp = g_strconcat(value, "}", 0);
        g_free(value);
        if (fwrite(tmp, sizeof(gchar), strlen(tmp), fp) < strlen(tmp)) {
            g_free(tmp);
            return -1;
        }
        g_free(tmp);
    }
    if (fwrite("\n", 1, 1, fp) < 1)
        return -1;        

    // Write fog color
    if (fwrite("fog_color=", 1, strlen("fog_color="), fp) < strlen("fog_color=")) {
        perror("Save options failed");
        return -1;
    }
    if (fwrite("{", 1, 1, fp) < 1)
        return -1;        
    for (int i = 0; i < 4; i++) {
        // .3 precision would be enough?
        value = g_strdup_printf("%.3f", fog_color[i]);
        if (!value)
            return -1;
        if (i < 3)
            tmp = g_strconcat(value, ",", 0);
        else
            tmp = g_strconcat(value, "}", 0);
        g_free(value);
        if (fwrite(tmp, sizeof(gchar), strlen(tmp), fp) < strlen(tmp)) {
            g_free(tmp);
            return -1;
        }
        g_free(tmp);
    }
    if (fwrite("\n", 1, 1, fp) < 1)
        return -1;    

    fclose(fp);

    return 0;
}


void Options::get_dir_color (GLfloat color[3])
{
    g_assert (options_readed);
    g_assert(color != NULL);
    for (int i = 0; i < 3; i++)
        color[i] = dir_color[i];
}


void Options::get_pdir_color (GLfloat color[3])
{
    g_assert (options_readed);
    g_assert(color != NULL);
    for (int i = 0; i < 3; i++)
        color[i] = pdir_color[i];
}


void Options::get_file_color (GLfloat color[3])
{
    g_assert (options_readed);
    g_assert(color != NULL);
    for (int i = 0; i < 3; i++)
        color[i] = file_color[i];
}


void Options::get_fog_color (GLfloat color[4])
{
    g_assert (options_readed);
    g_assert(color != NULL);
    for (int i = 0; i < 4; i++)
        color[i] = fog_color[i];
}


void Options::set_dir_color (const GLfloat color[3])
{
    g_assert(color != NULL);
    for (int i = 0; i < 3; i++)
        dir_color[i] = color[i];
}


void Options::set_pdir_color (const GLfloat color[3])
{
    g_assert(color != NULL);
    for (int i = 0; i < 3; i++)
        pdir_color[i] = color[i];
}


void Options::set_file_color (const GLfloat color[3])
{
    g_assert(color != NULL);
    for (int i = 0; i < 3; i++)
        file_color[i] = color[i];
}


void Options::set_fog_color (const GLfloat color[4])
{
    g_assert(color != NULL);
    for (int i = 0; i < 4; i++)
        fog_color[i] = color[i];
}



/*
 *
 * Private methods
 *
 */



/*
 * Parse the buffer, filling private Options
 * fields accordingly.
 *
 * buffer must point to a null terminated string.
 */
int Options::parse_options (const gchar *buffer, GHashTable *table)
{
    gchar *priv;
    gchar *ptr;
    gchar *line;
    gchar *name, *value;

    priv = g_strdup (buffer);
    ptr = priv;
    g_print("config options:\n");
    while (*ptr) {
        while (*ptr == '\n')
            ptr = g_utf8_next_char(ptr);
        // get the line
        line = ptr;
        ptr = g_utf8_strchr(ptr, g_utf8_strlen(ptr, -1), '\n');
        if (!ptr)
            break;
        // null-terminate the line
        *ptr = 0;
        // now ptr point to the next line, if any
        ptr = g_utf8_next_char(ptr);
        if (g_utf8_strlen(line, -1) > 1000) {
            g_warning("line %s on config file is too long", line);
            return -1;
        }
        // look for a comment line
        if (g_utf8_strchr(line, g_utf8_strlen(line, -1), '#') != NULL)
            continue;
        name = line;
        value = g_utf8_strchr(line, g_utf8_strlen(line, -1), '=');
        if (!value)
            g_error("config options error");
        // null-terminate name
        *value = 0;
        // skip '='
        value = g_utf8_next_char(value);
        // remove leading and trailing white spaces
        name = g_strstrip(name);
        value = g_strstrip(value);
        if ((!g_unichar_isgraph(name[0])) || (!g_unichar_isgraph(value[0])))
            g_error("config options error");
        g_print("\t%s = %s\n", name, value);
        g_hash_table_insert (table, name, value);
    }
    // FIXME if some option was not set, we must set a default
    return 0;
}


void Options::set_default_options ()
{
    for (int i = 0; i < 3; i++)
        dir_color[i] = default_dir_color[i];
    for (int i = 0; i < 3; i++)
        pdir_color[i] = default_pdir_color[i];
    for (int i = 0; i < 3; i++)
        file_color[i] = default_file_color[i];
    for (int i = 0; i < 4; i++)
        fog_color[i] = default_fog_color[i];
}


/*
 * Create a new config file.
 *
 * If config file already exists, it is overwritten.
 */
void Options::create_new_config_file ()
{
    gchar *config_dir;
    mode_t mode = S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
    FILE *fp;
    gchar *error_msg;

    config_dir = g_strconcat (g_get_home_dir(), G_DIR_SEPARATOR_S, _3DFB_CONFIG_DIR, 0);
    if (!g_file_test(config_dir, G_FILE_TEST_EXISTS)) {
        if (mkdir (config_dir, mode) < 0) {
            error_msg = g_strdup_printf("%s: config directory %s could not be created",
                                        PACKAGE_NAME, config_dir);
            perror(error_msg);
            g_free(error_msg);
            g_free(config_dir);
            return ;
        }
    }
    g_free(config_dir);
    fp = fopen(config_file, "w");
    if (fp == NULL) {
        error_msg = g_strdup_printf("%s: config file %s cannot be created",
                                    PACKAGE_NAME, config_file);
        perror(error_msg);
        g_free (error_msg);
    }
}


/*
 *
 * Auxiliary functions
 *
 */



static guint HashFunction (gconstpointer key)
{
    char *sKey;
    guint giHashValue;
    int nIndex;
    
    if (!key)
        return 0;
    giHashValue = 0;
    sKey = (char*) key;
    for (nIndex = 0; nIndex < (int)strlen(sKey); nIndex++)
        giHashValue = (giHashValue << 4) + (giHashValue ^ (guint) sKey[nIndex]);
    return giHashValue;
}


static gint HashCompare (gconstpointer sName1, gconstpointer sName2)
{
    // comparation is non case-sensitive
    return (!g_ascii_strcasecmp ((gchar*) sName1, (gchar*) sName2));
}


static void set_option_value (gpointer key, gpointer value, gpointer extra_data)
{
    gchar *name, *value_s;
    Options *opt;
    GLfloat c[3];

    name = (gchar*) key;
    value_s = (gchar*) value;
    opt = (Options*) extra_data;
    // comparation is non case-sensitive
    if (!g_ascii_strcasecmp(name, "dir_color")) {
        if (sscanf(value_s, "{%f,%f,%f}", &c[0], &c[1], &c[2]) != 3)
            g_print("Incorrect directory color format\n");
        else
            opt->set_dir_color(c);
    }
    else if (!g_ascii_strcasecmp(name, "pdir_color")) {
        if (sscanf(value_s, "{%f,%f,%f}", &c[0], &c[1], &c[2]) != 3)
            g_print("Incorrect parent directory color format\n");
        else
            opt->set_pdir_color(c);
    }
    else if (!g_ascii_strcasecmp(name, "file_color")) {
        if (sscanf(value_s, "{%f,%f,%f}", &c[0], &c[1], &c[2]) != 3)
            g_print("Incorrect file color format\n");
        else
            opt->set_file_color(c);
    }
    else if (!g_ascii_strcasecmp(name, "fog_color")) {
        GLfloat fc[4];
        
        if (sscanf(value_s, "{%f,%f,%f,%f}", &fc[0], &fc[1], &fc[2], &fc[3]) != 4)
            g_print("Incorrect fog color format\n");
        else
            opt->set_fog_color(fc);
    }
    else
        g_print("Unrecognized option %s\n", value_s);
}
